// serialServer.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
// Author: Marty Kraimer 
// Date: 7/98


#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include <epicsThread.h>
#include <epicsEvent.h>
#include <epicsTime.h>
#include <epicsRingBytes.h>
#include <epicsAssert.h>
#include <iocsh.h>
#include <epicsExport.h>

#include <asynDriver.h>
 
#include "Message.h"
#include "Int32Message.h"
#include "Char8ArrayMessage.h"
#include "SerialConfigMessage.h"
#include "serialServer.h"
 
class Serial;

extern "C" {
static void queueCallback(asynUser *pasynUser);
static void serialServerThread(Serial *pSerial);
}

class PortConfig {
public:
    PortConfig()
        : baud(9800), stopBits(1), bitsPerChar(8), parity('N') , flowControl('N')
        { }
    int32 baud;
    int16 stopBits;
    int16 bitsPerChar;
    char8 parity;
    char8 flowControl;
};
    
class Serial
{
public:
    Serial(const char *name,const char *portName, int addr,
	int bufsize, int queueSize, const char *eomstr);
    void processConfig(SerialConfigMessage *);
    void processGetConfig(Int32Message *);
    void processChar8Array(Char8ArrayMessage *);
    char *serverName;
    MessageServer *pMessageServer;
    bool startOK;
    asynStatus status;
    PortConfig config;
    char eomString[2];
    int eomLen;
    Message *message;
    epicsEventId done;
    int readBufSize;
    char *readBuf;
    asynUser *pasynUser;
    asynOctet *pasynOctet;
    void *asynOctetPvt;
    asynCommon *pasynCommon;
    void *asynCommonPvt;
};

static char taskname[] = "serial";
extern "C" int initSerialServer(const char *serverName,
    const char *portName, int addr, int bufsize, int queueSize, const char *eomstr)
{
    if(bufsize<=0) {
        printf("%s bufsize must be > 0\n",serverName);
        return(0);
    }
    Serial *pSerial =
	new Serial(serverName,portName,addr,bufsize,queueSize, eomstr);
    if(!pSerial->startOK) return(0);
    epicsThreadId threadId = epicsThreadCreate(
        taskname,epicsThreadPriorityMedium,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)serialServerThread,(void *)pSerial);
    if(!threadId)
	printf("%s serialServer epicsThreadCreate Failure\n",serverName);
    return(0);
}

Serial:: Serial(const char *name,const char *portName, int addr,
    int bufsize, int queueSize, const char *eomstr)
:   startOK(false),
    status(asynSuccess), readBufSize(0), readBuf(0)
{
    serverName = (char *)(new char[(size_t)name + 1]);
    strcpy(serverName,name);
    if(eomstr==0) {
	eomLen = 0;
    } else switch(strlen((const char  *)eomstr)) {
    case 0:
        eomString[0] = 0; eomLen = 1; break;
    case 1:
        eomString[0] = eomstr[0]; eomLen = 1; break;
    case 2:
        eomString[0] = eomstr[0]; eomString[1] = eomstr[1]; eomLen = 2; break;
    default:
	printf("%s serialServer illegal length eomstr. eomLen set to 0\n",
	    serverName);
	eomString[0] = 0; eomLen = 0; break;
    }
    readBuf = (char *)(new char[(size_t)bufsize]);
    readBufSize = bufsize;
    ::memset((void *)readBuf,'\0',(size_t)readBufSize);
    pasynUser = pasynManager->createAsynUser(queueCallback,0);
    pasynUser->userPvt = (void *)this;
    status = pasynManager->connectDevice(pasynUser,portName,addr);
    if(status!=asynSuccess) {
        printf("%s\n",pasynUser->errorMessage);
        pasynManager->freeAsynUser(pasynUser);
        return ;
    }
    asynInterface *pasynInterface;
    pasynInterface = pasynManager->findInterface(pasynUser,asynCommonType,1);
    if(!pasynInterface) {
        printf("%s %s\n",asynCommonType,pasynUser->errorMessage);
        pasynManager->freeAsynUser(pasynUser);
        return;
    }
    pasynCommon = (asynCommon *)pasynInterface->pinterface;
    asynCommonPvt = pasynInterface->drvPvt;
    pasynInterface = pasynManager->findInterface(
        pasynUser,asynOctetType,1);
    if(!pasynInterface) {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"%s %s\n",
            asynOctetType,pasynUser->errorMessage);
        pasynManager->freeAsynUser(pasynUser);
        return;
    }
    pasynOctet = (asynOctet *)pasynInterface->pinterface;
    asynOctetPvt = pasynInterface->drvPvt;
    done = epicsEventMustCreate(epicsEventEmpty);
    pMessageServer = new MessageServer(name,queueSize);
    startOK = true;
}

static void queueCallback(asynUser *pasynUser)
{
    Serial *pSerial = (Serial *)pasynUser->userPvt;
    Message *message = pSerial->message;

    switch(message->getType()) {
    case messageTypeSerialConfig:
	pSerial->processConfig((SerialConfigMessage *)message);
	break;
    case messageTypeInt32:
	pSerial->processGetConfig((Int32Message *)message);
	break;
    case messageTypeChar8Array:
	pSerial->processChar8Array((Char8ArrayMessage *)message);
	break;
    default:
        printf("%s serialServer got illegal message type %d\n",
            pSerial->serverName, message->getType());
        break;
    }
    epicsEventSignal(pSerial->done);
}

void serialServerThread(Serial *pSerial)
{
    asynUser *pasynUser = pSerial->pasynUser;
    const char *serverName = pSerial->serverName;
    asynStatus status;
    while(true) {
        pSerial->pMessageServer->waitForMessage();
	Message *message;
        while((message = pSerial->pMessageServer->receive())) {
            pSerial->message = message;
            asynPrint(pasynUser,ASYN_TRACE_FLOW,"%s calling lock\n",
                serverName);
            status =  pasynManager->lock(pasynUser);
            if(status!=asynSuccess) {
                asynPrint(pasynUser,ASYN_TRACE_ERROR,"%s lock error %s\n",
                    serverName,pasynUser->errorMessage);
            } 
            asynPrint(pasynUser,ASYN_TRACE_FLOW,
                "%s calling queueRequest\n",serverName);
            status = pasynManager->queueRequest(pasynUser,
               asynQueuePriorityLow,0.0);
            if(status!=asynSuccess) {
                asynPrint(pasynUser,ASYN_TRACE_ERROR,"%s %s\n",
                    serverName,pasynUser->errorMessage);
            } else {
                epicsEventMustWait(pSerial->done);
            }
            status =  pasynManager->unlock(pasynUser);
            if(status!=asynSuccess) {
                asynPrint(pasynUser,ASYN_TRACE_ERROR,"%s unlock error %s\n",
                    serverName,pasynUser->errorMessage);
            } 
	    delete message;
	}
    }
}

void Serial::processConfig(SerialConfigMessage *pconfig)
{
    PortConfig *pPortConfig = &this->config;
    Int32Message *preply = (Int32Message *)pMessageServer->
	allocReplyMessage(pconfig,messageTypeInt32);

    pPortConfig->baud = pconfig->baud;
    pPortConfig->stopBits = pconfig->stopBits;
    pPortConfig->bitsPerChar = pconfig->bitsPerChar;
    pPortConfig->parity = pconfig->parity;
    pPortConfig->flowControl = pconfig->flowControl;
    preply->status = asynError;
    pMessageServer->reply(preply);

    
}
void Serial::processGetConfig(Int32Message *pconfig)
{
    PortConfig *pPortConfig = &this->config;
    SerialConfigMessage *preply = (SerialConfigMessage *)pMessageServer->
	allocReplyMessage(pconfig,messageTypeSerialConfig);
    /* asynDriver does not implement getConfig. */
    preply->baud = pPortConfig->baud;
    preply->stopBits = pPortConfig->stopBits;
    preply->bitsPerChar = pPortConfig->bitsPerChar;
    preply->parity = pPortConfig->parity;
    preply->flowControl = pPortConfig->flowControl;
    pMessageServer->reply(preply);
}

void Serial::processChar8Array(Char8ArrayMessage *preceive)
{
    int cmd = preceive->cmd;
    int nchars = 0;
    int maxchars;

    if(preceive->timeoutUnits==timeoutUnitsMilliseconds) {
        pasynUser->timeout = (double)(preceive->timeout/1000.0);
    } else {
        pasynUser->timeout = (double)(preceive->timeout);
    }
    Char8ArrayMessage *preply = (Char8ArrayMessage *)pMessageServer->
        allocReplyMessage(preceive,messageTypeChar8Array);
    if(cmd&cmdSetEom) {
        eomLen = preceive->eomLen;
        eomString[0] = preceive->eomString[0];
        eomString[1] = preceive->eomString[1];
    }
    asynPrint(pasynUser,ASYN_TRACE_FLOW,"%s setEos eoslen %d\n",
        serverName,eomLen);
    status = pasynOctet->setEos(asynOctetPvt,pasynUser,eomString,eomLen);
    if(status!=asynSuccess)  asynPrint(pasynUser,ASYN_TRACE_ERROR,
        "%s setEos error %s\n",serverName,pasynUser->errorMessage);
    if(cmd&cmdFlush) {
        asynPrint(pasynUser,ASYN_TRACE_FLOW,"%s flush\n",serverName);
        status = pasynOctet->flush(asynOctetPvt,pasynUser);
        if(status!=asynSuccess)  asynPrint(pasynUser,ASYN_TRACE_ERROR,
            "%s flush error %s\n",serverName,pasynUser->errorMessage);
    }
    if(cmd&cmdWrite) {
        const char *data = (const char *)preceive->value;
        nchars = preceive->getSize();
        asynPrint(pasynUser,ASYN_TRACE_FLOW,"%s write\n",serverName);
        asynPrintIO(pasynUser,ASYN_TRACEIO_DEVICE,data,nchars,
            "%s write %d chars ",serverName,nchars);
        int nout = pasynOctet->write(asynOctetPvt,pasynUser,
            (const char *)preceive->value,preceive->getSize());
        if(nout != nchars) {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,
            "%s write error %s\n",serverName,pasynUser->errorMessage);
        }
    }
    if(status!=asynSuccess || !(cmd&cmdRead)) {
        preply->status = status;
        pMessageServer->reply(preply);
        return;
    }
    assert(cmd&cmdRead);
    maxchars = (preceive->extra>0) ? preceive->extra : readBufSize;
    asynPrint(pasynUser,ASYN_TRACE_FLOW,"%s read\n",serverName);
    nchars = pasynOctet->read(asynOctetPvt,pasynUser,readBuf,maxchars);
    asynPrintIO(pasynUser,ASYN_TRACEIO_DEVICE,readBuf,nchars,
        "%s read %d chars ",serverName,nchars);
    if(nchars<=0) {
        preply->status = asynTimeout;;
        pMessageServer->reply(preply);
        return;
    }
    preply->status = asynSuccess;
    preply->allocValue(nchars);
    ::memcpy((void *)preply->value,(void *)readBuf,nchars);
    preply->setSize(nchars);
    pMessageServer->reply(preply);
    return;
}

extern "C" {

static const iocshArg initSerialServerArg0 = { "serverName",iocshArgString};
static const iocshArg initSerialServerArg1 = { "portname",iocshArgString};
static const iocshArg initSerialServerArg2 = { "addr",iocshArgInt};
static const iocshArg initSerialServerArg3 = { "bufSize",iocshArgInt};
static const iocshArg initSerialServerArg4 = { "queueSize",iocshArgInt};
static const iocshArg initSerialServerArg5 = { "eomString",iocshArgString};
static const iocshArg * const initSerialServerArgs[] = {
    &initSerialServerArg0,&initSerialServerArg1,&initSerialServerArg2,
    &initSerialServerArg3,&initSerialServerArg4,&initSerialServerArg5};
static const iocshFuncDef initSerialServerFuncDef = {
    "initSerialServer",6,initSerialServerArgs};
static void initSerialServerCallFunc(const iocshArgBuf *args)
{
    initSerialServer(
        args[0].sval,args[1].sval,args[2].ival,
        args[3].ival,args[4].ival,args[5].sval);
}

static void serialServerRegister(void)
{
    iocshRegister(&initSerialServerFuncDef,initSerialServerCallFunc);
}
epicsExportRegistrar(serialServerRegister);
} // extern "C"
